import java.io.ByteArrayInputStream;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.Provider;
import java.security.Security;
import java.security.Signature;
import java.security.Principal;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import javax.security.auth.callback.CallbackHandler;
import sun.security.pkcs11.SunPKCS11;
import java.io.IOException;
import java.io.OutputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import javax.net.ssl.X509KeyManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.SSLSocket;
import java.net.Socket;
import java.util.Enumeration;


public class TLSClientAuth {

	static Provider pkcs11_provider;

	private static String getPKCS11Library() {
		final String LIBNAME = "justica-pkcs11.dll";
		final String WIN_FOLDER = "C:\\Windows\\System32\\", SYSWOW_FOLDER = "C:\\Windows\\SysWOW64\\";

		String os = System.getProperty("os.name").toLowerCase();

		if (os.contains("windows")) {
			String process_arch = System.getProperty("sun.arch.data.model");
			
			if (process_arch.equals("32"))
			{
				if (System.getenv("ProgramFiles(x86)") != null) //32-bit JVM on Win64
				{
					return SYSWOW_FOLDER + LIBNAME; 
				}
				else
					return WIN_FOLDER + LIBNAME;      //32-bit JVM on Win32

			}
			else return WIN_FOLDER + LIBNAME;       //64-bit JVM on Win64

		}
		else if (os.contains("linux")) {
			return "/usr/local/lib/libjustica-pkcs11.so";
		}
		else if (os.contains("mac")) {
			return "/usr/local/lib/libjustica-pkcs11.dylib";
		}
		else {
			throw new UnsupportedOperationException("Unknown or unsupported operating system");
		}
	}

	static {

		String pkcs11config = "name=JusticaPKCS11" + "\n"+ "library="+ getPKCS11Library();

		byte[] pkcs11configBytes = pkcs11config.getBytes();
		ByteArrayInputStream configStream = new ByteArrayInputStream(pkcs11configBytes);

		pkcs11_provider = new sun.security.pkcs11.SunPKCS11(configStream);
		
		Security.addProvider(pkcs11_provider);

	}


	public static void main(String[] args) {

		final String hostname = "ubuntu"; 
		int port = 443;

		try {

			CallbackHandler cmdLineHdlr = new com.sun.security.auth.callback.TextCallbackHandler();
			KeyStore.Builder builder = KeyStore.Builder.newInstance("PKCS11", pkcs11_provider,
					    new KeyStore.CallbackHandlerProtection(cmdLineHdlr));
			KeyStore ks = builder.getKeyStore();
			ks.load(null, null);

			/*
			System.out.println("List of keystore entries: ");
			for (Enumeration<String> alias = ks.aliases(); alias.hasMoreElements();)
       				System.out.println(alias.nextElement());
       		*/		

			//Initialize key and trust Managers 
			KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
			keyManagerFactory.init(ks, null);

			final X509KeyManager origKm = (X509KeyManager)keyManagerFactory.getKeyManagers()[0];

			X509KeyManager km = new X509KeyManager() {
    			public String chooseClientAlias(String[] keyType, Principal[] issuers, Socket socket) {
        			return "AUTHENTICATION CERTIFICATE";
    			}

    			public String chooseServerAlias(String keyType, Principal[] issuers, Socket socket) {
        			return origKm.chooseServerAlias(keyType, issuers, socket);
    			}

    			public X509Certificate[] getCertificateChain(String alias) {
        			return origKm.getCertificateChain(alias);
    			}

    			public PrivateKey getPrivateKey(String alias) {
    				return origKm.getPrivateKey(alias);
    			}
    			public String[] getClientAliases(String keyType, Principal[] issuers) {
    				return origKm.getClientAliases(keyType, issuers);
    			}
    			public String[] getServerAliases(String keyType, Principal[] issuers) {
    				return origKm.getServerAliases(keyType, issuers);
    			}
			};

			TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
			trustManagerFactory.init((KeyStore)null);

			//Initialize TLS context
			SSLContext context = SSLContext.getInstance("TLS");
			context.init(new KeyManager[] { km }, trustManagerFactory.getTrustManagers(), null);

			//Create client socket
			
			SSLSocketFactory ssf = context.getSocketFactory();

			SSLSocket s = (SSLSocket) ssf.createSocket(hostname, port);
			s.setUseClientMode(true);
			s.setSoTimeout(30000);
			s.startHandshake();

			OutputStream os = s.getOutputStream();
			String request = "GET /SP HTTP/1.1\r\nHost: "+ hostname + "\r\n\r\n";
			os.write(request.getBytes());

			BufferedReader br = new BufferedReader(new InputStreamReader(s.getInputStream()));
			String t;
			while((t = br.readLine()) != null) 
				System.out.println(t);
			br.close();

		}
		catch (KeyStoreException e) {
			System.err.println("Keystore exception thrown!");
			e.printStackTrace();
			System.exit(1);
		}
		catch(IOException e) {
			System.err.println("I/O exception thrown!");
			e.printStackTrace();
			System.exit(1);
		}
		catch (Exception e) {
			System.err.println("General exception thrown!");
			e.printStackTrace();	
		}

	}


}